from typing import Any, Dict, Optional

import orm

from aopi_python.ctx import context
from aopi_python.models.dist_info import DistInfoModel


class PythonPackage(orm.Model):
    __tablename__ = "python_packages"
    __database__ = context.database
    __metadata__ = context.metadata

    id = orm.Integer(primary_key=True)
    user_id = orm.Integer(allow_null=True, index=True)

    name = orm.Text(allow_null=False, index=True)
    author = orm.Text(allow_null=True)
    summary = orm.Text(allow_null=True)
    license = orm.Text(allow_null=True)
    keywords = orm.Text(allow_null=True)
    home_page = orm.Text(allow_null=True)
    maintainer = orm.Text(allow_null=True)
    project_urls = orm.Text(allow_null=True)
    author_email = orm.Text(allow_null=True)
    maintainer_email = orm.Text(allow_null=True)

    @staticmethod
    def cast_dist_info_to_data(upload: DistInfoModel) -> Dict[str, Any]:
        return dict(
            name=upload.name,
            summary=upload.summary,
            license=upload.license,
            keywords=upload.keywords,
            home_page=upload.home_page,
            author=upload.author,
            author_email=upload.author_email,
            project_urls=";".join(upload.project_urls) if upload.project_urls else None,
            maintainer=upload.maintainer,
            maintainer_email=upload.maintainer_email,
        )

    @classmethod
    async def create_by_dist_info(
        cls, *, upload: DistInfoModel, user_id: Optional[int]
    ) -> "PythonPackage":
        upload_dict = cls.cast_dist_info_to_data(upload)
        return await cls.objects.create(**upload_dict, user_id=user_id)

    async def update_by_dist_info(self, upload: DistInfoModel) -> None:
        upload_dict = self.cast_dist_info_to_data(upload)
        await self.update(**upload_dict)



from typing import Dict, List, Optional

from aopi_index_builder import FullPackageInfo, PackagePreview, PackageVersion
from aopi_index_builder.schema import ReadmeFormats

from aopi_python.ctx import context
from aopi_python.models import PythonPackage, PythonPackageVersion
from aopi_python.roles import RolesEnum


async def find_packages_func(
    user_id: Optional[int], pkg_name: str, limit: int, offset: int
) -> List[PackagePreview]:
    await context.has_permission(user_id=user_id, role=RolesEnum.read)

    def to_preview(package: Dict[str, str]) -> PackagePreview:
        return PackagePreview(
            id=package["id"], name=package["name"], short_description=package["summary"]
        )

    expr = PythonPackage.objects.filter(
        name__icontains=pkg_name
    ).build_select_expression()
    packages = await context.database.fetch_all(expr.limit(limit).offset(offset))
    return list(map(to_preview, packages))


async def get_package_info_func(user_id: Optional[int], pkg_id: int) -> FullPackageInfo:
    await context.has_permission(user_id=user_id, role=RolesEnum.read)
    package: PythonPackage = await PythonPackage.objects.get(id=pkg_id)
    last_version_query = PythonPackageVersion.objects.filter(
        package=package
    ).build_select_expression()
    last_version_query.order_by(
        PythonPackageVersion.__table__.c.upload_time.desc()
    ).limit(1)
    last_version = await context.database.fetch_one(last_version_query)
    description = None
    description_format = ReadmeFormats.TEXT
    if "description" in last_version.keys():
        description = last_version["description"]
    content_type_map = {
        "text/x-rst": ReadmeFormats.RST,
        "text/markdown": ReadmeFormats.MD,
    }
    if "description_content_type" in last_version.keys():
        description_format = content_type_map.get(
            last_version["description_content_type"], description_format
        )
    return FullPackageInfo(
        id=package.id,
        name=package.name,
        short_description=package.summary,
        description=description,
        description_format=description_format,
        last_version=last_version["version"],
        metadata={
            key: val
            for key, val in package.items()
            if key not in ["id", "name", "summary", "description"]
        },
    )


async def get_versions_func(
    user_id: Optional[int], pkg_id: int
) -> List[PackageVersion]:
    await context.has_permission(user_id=user_id, role=RolesEnum.read)

    def to_package_version(version: PythonPackageVersion) -> PackageVersion:
        return PackageVersion(
            version=version.version,
            yanked=version.yanked,
            metadata={
                key: val
                for key, val in version.items()
                if key
                not in [
                    "id",
                    "yanked",
                    "version",
                    "user_id",
                    "package",
                    "description",
                ]
            },
        )

    versions = await PythonPackageVersion.objects.filter(package__id=pkg_id).all()

    return list(map(to_package_version, versions))



from enum import Enum, unique
from typing import Any, Dict, List, Optional

from fastapi import File, UploadFile
from pydantic import Field
from pydantic.main import BaseConfig, BaseModel

from aopi_python.utils import as_form


@unique
class ReadmeContentType(str, Enum):
    MD = "text/markdown"
    RST = "text/x-rst"
    PLAIN = "text/plain"


class DistInfoModel(BaseModel):
    name: str
    version: str
    filetype: str
    metadata_version: float
    md5_digest: Optional[str]
    sha256_digest: Optional[str]
    requires_python: Optional[str]
    protocol_version: Optional[str]
    author: Optional[str]
    summary: Optional[str]
    blake2_256_digest: Optional[str]
    comment: Optional[str]
    license: Optional[str]
    keywords: Optional[str]
    provides: Optional[str]
    requires: Optional[str]
    obsoletes: Optional[str]
    home_page: Optional[str]
    maintainer: Optional[str]
    description: Optional[str]
    author_email: Optional[str]
    download_url: Optional[str]
    provides_dist: Optional[str]
    platform: Optional[List[str]]
    obsoletes_dist: Optional[str]
    maintainer_email: Optional[str]
    classifiers: Optional[List[str]]
    requires_external: Optional[str]
    project_urls: Optional[List[str]]
    requires_dist: Optional[List[str]]
    supported_platform: Optional[List[str]]
    description_content_type: Optional[ReadmeContentType]
    python_version: Optional[str] = Field(None, alias="pyversion")


@as_form
class PackageUploadModel(DistInfoModel):
    action: str = Field(..., alias=":action")
    content: UploadFile = File(...)

    class Config(BaseConfig):
        arbitrary_types_allowed = True
        allow_population_by_field_name = True

    @staticmethod
    def as_form(**data: Dict[str, Any]) -> "PackageUploadModel":
        ...



import inspect
from typing import Any, Dict, Type

from fastapi import File, Form
from pydantic import BaseModel
from pydantic.fields import ModelField


def as_form(cls: Type[BaseModel]) -> Type[BaseModel]:
    new_parameters = []

    for field_name, model_field in cls.__fields__.items():
        model_field: ModelField  # type: ignore

        def_val = None
        if model_field.required:
            def_val = model_field.default or ...

        form_default = Form(def_val, alias=model_field.alias)
        if model_field.default.__class__ is File:
            form_default = model_field.default

        new_parameters.append(
            inspect.Parameter(
                model_field.name,
                inspect.Parameter.POSITIONAL_ONLY,
                default=form_default,
                annotation=model_field.outer_type_,
            )
        )

    async def as_form_func(**data: Dict[str, Any]) -> BaseModel:
        return cls(**data)

    sig = inspect.signature(as_form_func)
    sig = sig.replace(parameters=new_parameters)
    as_form_func.__signature__ = sig  # type: ignore
    setattr(cls, "as_form", as_form_func)
    return cls



from datetime import datetime
from typing import Any, Dict, Optional

import orm

from aopi_python.ctx import context
from aopi_python.models.dist_info import DistInfoModel
from aopi_python.models.pythonpackage import PythonPackage


class PythonPackageVersion(orm.Model):
    __tablename__ = "python_packages_versions"
    __database__ = context.database
    __metadata__ = context.metadata

    id = orm.Integer(primary_key=True)
    package = orm.ForeignKey(PythonPackage)
    user_id = orm.Integer(index=True, allow_null=True)

    size = orm.Integer()
    version = orm.Text()
    requires_python = orm.Text()
    yanked = orm.Boolean(default=False)
    comment_text = orm.Text(allow_null=True)
    yanked_reason = orm.Text(allow_null=True)
    filetype = orm.String(max_length=30)
    metadata_version = orm.Float()
    downloads = orm.Integer(default=0)
    filename = orm.Text(allow_null=False)
    md5_digest = orm.Text(allow_null=True)
    description = orm.Text(allow_null=True)
    sha256_digest = orm.Text(allow_null=True)
    upload_time = orm.DateTime(default=datetime.now)
    requires_dist = orm.Text(allow_null=True)
    description_content_type = orm.Text(allow_null=True)
    python_version = orm.Text(allow_null=True)

    @staticmethod
    def cast_upload_to_dict(filename: str, upload: DistInfoModel) -> Dict[str, Any]:
        return dict(
            version=upload.version,
            description=upload.description,
            comment_text=upload.comment,
            filetype=upload.filetype,
            metadata_version=upload.metadata_version,
            filename=filename,
            md5_digest=upload.md5_digest,
            requires_python=upload.requires_python,
            sha256_digest=upload.sha256_digest,
            description_content_type=upload.description_content_type.value
            if upload.description_content_type
            else None,
            requires_dist=";".join(upload.requires_dist)
            if upload.requires_dist
            else None,
            python_version=upload.python_version,
        )

    @classmethod
    async def create_by_dist_info(
        cls,
        *,
        filename: str,
        package: PythonPackage,
        size: int,
        dist_info: DistInfoModel,
        user_id: Optional[int]
    ) -> "PythonPackageVersion":
        info_dict = cls.cast_upload_to_dict(filename, dist_info)
        return await cls.objects.create(
            package=package, size=size, user_id=user_id, **info_dict
        )

    async def update_by_dist_info(
        self,
        *,
        filename: str,
        package: PythonPackage,
        size: int,
        dist_info: DistInfoModel
    ) -> "PythonPackageVersion":
        info_dict = self.cast_upload_to_dict(filename, dist_info)
        return await self.update(package=package, size=size, **info_dict)



import shutil
from operator import attrgetter, itemgetter
from typing import List, Optional

import sqlalchemy
from fastapi import APIRouter, Depends, HTTPException
from loguru import logger
from natsort import natsorted
from orm import NoMatch
from sqlalchemy.sql import Select
from starlette.requests import Request
from starlette.responses import HTMLResponse, Response

from aopi_python import models
from aopi_python.ctx import context, templates
from aopi_python.models.dist_info import PackageUploadModel
from aopi_python.roles import RolesEnum
from aopi_python.routes.simple.dependencies import get_user_with_role
from aopi_python.routes.simple.logic import save_file

simple_router = APIRouter(prefix="/simple")
PREFIX = f"{context.prefix}{simple_router.prefix}"


@simple_router.get("", response_class=HTMLResponse)
async def python_simple_index_page(
    request: Request, _: Optional[int] = Depends(get_user_with_role(RolesEnum.read))
) -> templates.TemplateResponse:
    select: Select = sqlalchemy.sql.select([models.PythonPackage.objects.table.c.name])
    packages = map(itemgetter(0), await context.database.fetch_all(select))
    return templates.TemplateResponse(
        "simple/index.jinja2",
        {
            "prefix": PREFIX,
            "packages": packages,
            "request": request,
        },
    )


@simple_router.post("")
async def upload_python_package(
    response: Response,
    upload: PackageUploadModel = Depends(PackageUploadModel.as_form),
    user_id: Optional[int] = Depends(get_user_with_role(RolesEnum.upload)),
) -> Response:
    logger.debug(f"Trying to upload {upload.name}:{upload.version}")
    pkg_dir = context.packages_dir / upload.name / upload.version / upload.filetype
    try:
        await models.PythonPackageVersion.objects.get(
            package__name=upload.name,
            version=upload.version,
            filetype=upload.filetype,
        )
        logger.debug(f"Failed to upload {upload.name}. Already exists.")
        raise HTTPException(status_code=409, detail="Distribution already exists.")
    except NoMatch:
        pass
    package_exists = await models.PythonPackage.objects.filter(
        name=upload.name
    ).exists()
    try:
        file_path = pkg_dir / upload.content.filename
        await save_file(file_path, upload.content)
        logger.debug(f"Saved package file {upload.name} {upload.filetype}")
        if package_exists:
            package = await models.PythonPackage.objects.get(name=upload.name)
            await package.update_by_dist_info(upload)
        else:
            package = await models.PythonPackage.create_by_dist_info(
                upload=upload, user_id=user_id
            )
        if user_id != package.user_id:
            raise HTTPException(status_code=403, detail="You aren't the author")
        await models.PythonPackageVersion.create_by_dist_info(
            filename=upload.content.filename,
            package=package,
            size=file_path.stat().st_size,
            dist_info=upload,
            user_id=user_id,
        )
    except Exception as e:
        logger.exception(e)
        if pkg_dir.exists():
            shutil.rmtree(pkg_dir)
        raise HTTPException(
            status_code=400, detail="Something went wrong during saving."
        )

    response.status_code = 201
    return response


@simple_router.get("/{pkg_name}/", response_class=HTMLResponse)
async def get_python_package_info(
    pkg_name: str,
    request: Request,
    _: Optional[int] = Depends(get_user_with_role(RolesEnum.read)),
) -> templates.TemplateResponse:
    """
    Get information about the package.

    """
    versions: List[
        models.PythonPackageVersion
    ] = await models.PythonPackageVersion.objects.filter(package__name=pkg_name).all()
    if not versions:
        raise HTTPException(status_code=404, detail="Package was not found")
    versions = natsorted(versions, key=attrgetter("version"))
    readme = str() if not versions else versions[-1].description
    return templates.TemplateResponse(
        "simple/package_versions.jinja2",
        {
            "request": request,
            "name": pkg_name,
            "prefix": PREFIX,
            "versions": versions,
            "readme": readme,
        },
    )



import os
from pathlib import Path
from typing import Optional

import ujson
from aiofile import async_open
from fastapi import UploadFile
from loguru import logger

from aopi_python.models.dist_info import PackageUploadModel


async def update_info(version_dir: Path, upload: PackageUploadModel) -> None:
    info_path = version_dir / "info.json"
    info_dict = {}
    if info_path.exists():
        async with async_open(info_path, "r") as f:
            info_dict = ujson.loads(await f.read())
    dist_info = upload.dict(exclude={"content"})
    dist_info["filename"] = str(upload.content.filename)
    info_dict[upload.filetype] = dist_info
    async with async_open(info_path, "w") as f:
        await f.write(ujson.dumps(info_dict))


async def update_readme(version_dir: Path, description: Optional[str]) -> None:
    if description is None:
        return
    readme = version_dir / "README"
    if readme.exists():
        async with async_open(readme, "r") as f:
            readme_text = await f.read()
        if readme_text == description:
            return
    async with async_open(readme, "w") as f:
        await f.write(description)


async def save_file(file_path: Path, target_file: UploadFile) -> None:
    target_dir = os.path.dirname(file_path)
    if not os.path.exists(target_dir):
        os.makedirs(target_dir)
        logger.debug(file_path)
    await target_file.seek(0)
    async with async_open(file_path, "wb") as dest_file:
        while chunk := await target_file.read(1000):
            await dest_file.write(chunk)



from fastapi import APIRouter

from aopi_python.routes.file_router import file_router
from aopi_python.routes.simple import simple_router

main_router = APIRouter()

main_router.include_router(simple_router)
main_router.include_router(file_router)



from aopi_index_builder import PackageIndex

from aopi_python.ctx import context
from aopi_python.index_functions import (
    find_packages_func,
    get_package_info_func,
    get_versions_func,
)
from aopi_python.models import models_list
from aopi_python.roles import RolesEnum
from aopi_python.routes import main_router


def main() -> PackageIndex:
    return PackageIndex(
        router=main_router,
        target_language="python",
        db_models=models_list,
        roles=RolesEnum,
        plugin_help=f"To use it add {{aopi-url}}/"
        f"{context.prefix}/simple as your index-url",
        find_packages_func=find_packages_func,
        get_package_info_func=get_package_info_func,
        get_versions_func=get_versions_func,
    )



from pathlib import Path

from aopi_index_builder import get_context
from starlette.templating import Jinja2Templates

context = get_context()
plugin_dir = Path(__file__).parent
templates_dir = plugin_dir / "templates"
templates = Jinja2Templates(directory=templates_dir)



from typing import Awaitable, Callable, Optional

from fastapi import Depends
from fastapi.security import HTTPBasic, HTTPBasicCredentials
from starlette.requests import Request

from aopi_python.ctx import context
from aopi_python.roles import RolesEnum

if context.enable_users:
    security = HTTPBasic()
else:

    async def security(_request: Request) -> None:
        return None


def get_user_with_role(
    role: RolesEnum,
) -> Callable[[Optional[HTTPBasicCredentials]], Awaitable[Optional[int]]]:
    async def route_security(
        credentials: Optional[HTTPBasicCredentials] = Depends(security),
    ) -> Optional[int]:
        if credentials is None:
            return None
        user_id = await context.get_user_id(credentials.username, credentials.password)
        await context.has_permission(user_id, role)
        return user_id

    return route_security



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aopi_python',
 'aopi_python.models',
 'aopi_python.routes',
 'aopi_python.routes.simple',
 'aopi_python.utils']

package_data = \
{'': ['*'], 'aopi_python': ['templates/simple/*']}

install_requires = \
['aiofile>=3.3.3,<4.0.0',
 'aopi-index-builder>=0.1.40,<0.2.0',
 'loguru>=0.5.3,<0.6.0',
 'natsort>=7.1.0,<8.0.0',
 'orm>=0.1.5,<0.2.0',
 'ujson>=4.0.1,<5.0.0']

entry_points = \
{'aopi_index': ['python = aopi_python.main:main']}

setup_kwargs = {
    'name': 'aopi-python',
    'version': '0.1.30',
    'description': '',
    'long_description': 'Python package index module.\n============================\n\nThis module adds python package index to your aopi server.',
    'author': 'Pavel Kirilin',
    'author_email': 'win10@list.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from typing import Optional

from fastapi import APIRouter, Depends
from starlette.responses import FileResponse

from aopi_python.ctx import context
from aopi_python.roles import RolesEnum
from aopi_python.routes.simple.dependencies import get_user_with_role

file_router = APIRouter()


@file_router.get("/python_files/{pkg_name}/{version}/{dist_type}/{filename}")
async def get_python_dist_file(
    pkg_name: str,
    version: str,
    dist_type: str,
    filename: str,
    _: Optional[int] = Depends(get_user_with_role(RolesEnum.read)),
) -> FileResponse:
    file_path = context.packages_dir / pkg_name / version / dist_type / filename
    return FileResponse(
        file_path,
        status_code=200,
        headers={"Content-Disposition": f"attachment; filename={filename}"},
    )
